import re
from base64 import b64decode
from typing import List

import requests
from requests import PreparedRequest

from .authenticator import Authenticator
from .context import MemfaultCliClickContext

EXPORTED_CHUNKS_REGEX = re.compile(r"MC:([a-zA-z0-9+/=]+):")


class MemfaultChunk:
    def __init__(self, ctx: MemfaultCliClickContext, authenticator: Authenticator):
        self.ctx = ctx
        self.authenticator = authenticator

    @property
    def url(self):
        return f"{self.ctx.chunks_url}/api/v0/chunks/{self.ctx.device_serial}"

    def post(self, chunk: bytes):
        request_args = self.authenticator.requests_auth_params()
        request_args["headers"]["Content-Type"] = "application/octet-stream"
        response = requests.post(self.url, data=chunk, **request_args)
        if response.status_code >= 400:
            raise Exception(
                f"Request failed with HTTP status {response.status_code}\nResponse body:\n{response.content.decode()}"
            )

    def batch_post(self, chunks: List[bytes]):
        if len(chunks) == 1:
            self.post(chunks[0])
            return

        files = []
        for idx, chunk in enumerate(chunks):
            files.append(
                (
                    # "Content-Disposition" header
                    f"chunk{idx}",
                    # (filename, fileobj, contentype, custom_headers)
                    (
                        f"chunk.bin{idx}",
                        chunk,
                        "application/octet-stream",
                        {"Content-Length": len(chunk)},
                    ),
                )
            )

        request_args = self.authenticator.requests_auth_params()

        # Pick up defaults (i.e User-Agent) and merge with Memfault Auth headers
        headers = requests.utils.default_headers()
        headers.update(request_args["headers"])
        del request_args["headers"]

        request = PreparedRequest()
        request.prepare(method="POST", url=self.url, files=files, headers=headers, **request_args)
        content_type: str = request.headers["content-type"]
        request.headers["content-type"] = content_type.replace("form-data", "mixed")

        session = requests.Session()
        response = session.send(request)
        if response.status_code >= 400:
            raise Exception(
                f"Request failed with HTTP status {response.status_code}\nResponse body:\n{response.content.decode()}"
            )

    @staticmethod
    def extract_exported_chunks(data: str):
        chunks = []
        for match in EXPORTED_CHUNKS_REGEX.finditer(data):
            binary_encoded_chunk = b64decode(match.group(1))
            chunks.append(binary_encoded_chunk)
        return chunks



import abc
import hashlib
import logging
import os
import re
import urllib.parse
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from glob import glob
from typing import IO, Dict, Iterable, List, Optional, Tuple, Type, Union
from zipfile import is_zipfile

import click
import requests
from elftools.elf.elffile import ELFFile
from mflt_build_id import BuildIdInspectorAndPatcher
from requests import Response, Session
from requests.adapters import HTTPAdapter
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper
from urllib3 import Retry

from .authenticator import (
    Authenticator,
    BasicAuthenticator,
    OrgTokenAuthenticator,
    ProjectKeyAuthenticator,
)
from .context import DeltaReleaseVersion, MemfaultCliClickContext, PackageInfo
from .elf import elf_has_debug_info, get_gnu_build_id, is_elf
from .functools_ext import cached_property

LOG = logging.getLogger(__name__)

GNU_BUILD_ID_RE = re.compile(r"Build ID: (?P<gnu_build_id>[a-f\d]+)")


def hash_io_hexdigest(name, data: IO, chunksize=2**20) -> str:
    """
    A wrapper around hashlib.new() and hexdigest() which reads
    from a stream in chunks
    """
    assert data.readable()

    # In case it's a large file, read it in chunks.
    h = hashlib.new(name)
    while True:
        chunk = data.read(chunksize)
        if not chunk:
            break
        h.update(chunk)
    return h.hexdigest()


def check_response(response: Response):
    if response.status_code >= 400:
        raise Exception(
            f"Request failed with HTTP status {response.status_code}\n"
            f"Response body:\n{response.content.decode() if response.content else '(empty)'}"
        )


def url_quote(value: str) -> str:
    return urllib.parse.quote(value, safe="", encoding="utf-8")


class Uploader(metaclass=abc.ABCMeta):
    authenticator_types: List[Type[Authenticator]]

    def __init__(
        self,
        *,
        ctx: MemfaultCliClickContext,
        file_path: Union[str, os.PathLike],
        authenticator: Authenticator,
        **kwargs,
    ):
        self.ctx: MemfaultCliClickContext = ctx
        self.file_path = file_path
        self.authenticator: Authenticator = authenticator
        self.session = self._create_requests_session()

    @staticmethod
    def _create_requests_session() -> Session:
        retry_strategy = Retry(
            total=5,
            backoff_factor=2,  # Sleep for 2s, 4s, 8s, 16s, 32s, Stop
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "POST", "PUT"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        return session

    @property
    def base_url(self):
        return self.ctx.file_url

    @property
    def app_url(self):
        return self.ctx.app_url

    def _api_base_url(self) -> str:
        return f"{self.base_url}/api/v0"

    def _projects_base_url(self, *, base_url_override=None) -> str:
        base_url = base_url_override if base_url_override is not None else self.base_url
        return f"{base_url}/api/v0/organizations/{self.ctx.org}/projects/{self.ctx.project}"

    def _projects_ui_base_url(self) -> str:
        return f"{self.app_url}/organizations/{self.ctx.org}/projects/{self.ctx.project}"

    def upload_url(self) -> str:
        """
        The upload URL for a 'prepared upload' for the configured authentication method.
        """
        if self.authenticator.project_key_auth():
            return f"{self._api_base_url()}/upload"
        else:
            return f"{self._projects_base_url()}/upload"

    @abstractmethod
    def can_upload_file(self) -> bool:
        """Test the file pointed to by `file_path`."""
        pass

    @abstractmethod
    def entity_url(self) -> str:
        """The final URL to POST to during the prepared upload sequence."""
        pass

    def ui_url(self) -> Optional[str]:
        """The UI URL at which results from the operation can be seen"""
        return None

    def _is_already_uploaded(self) -> bool:
        return False

    def _prepare_upload(self) -> Tuple[str, str]:
        response = self.session.post(self.upload_url(), **self.authenticator.requests_auth_params())
        check_response(response)
        data = response.json()["data"]
        return data["upload_url"], data["token"]

    def _put_file(self, upload_url: str, *, progressbar=True) -> None:
        file_size = os.stat(self.file_path).st_size
        with open(self.file_path, "rb") as file:
            with tqdm(
                total=file_size,
                desc=os.path.basename(file.name),
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                disable=not progressbar,
            ) as t:
                wrapped_file = CallbackIOWrapper(t.update, file, "read")
                response = self.session.put(upload_url, data=wrapped_file)
                check_response(response)

    def post_token_extra_request_data(self) -> Optional[Dict]:
        return None

    def _post_token(self, token: str) -> None:
        json_d = {
            "file": {"token": token, "md5": self.file_md5, "name": self.file_basename},
        }
        extra_request_data = self.post_token_extra_request_data()
        if extra_request_data:
            json_d.update(extra_request_data)

        response = self.session.post(
            self.entity_url(), json=json_d, **self.authenticator.requests_auth_params()
        )
        # This file had already uploaded (maybe concurrently). Gracefully ignore the error then.
        if response.status_code == 409:
            LOG.info("%s: was already uploaded.", self.file_path)
            return

        check_response(response)

    @property
    def qcomm_project(self):
        response = self.session.get(
            self._projects_base_url(base_url_override=self.ctx.api_url),
            **self.authenticator.requests_auth_params(),
        )
        check_response(response)
        return response.json()["data"].get("qcomm_enabled", False)

    def upload(self, *, progressbar=True) -> bool:
        if self._is_already_uploaded():
            LOG.info("%s: skipping, already uploaded.", self.file_path)
            return False
        upload_url, token = self._prepare_upload()
        self._put_file(upload_url, progressbar=progressbar)
        self._post_token(token)
        LOG.info("%s: uploaded!", self.file_path)
        if self.ui_url():
            click.echo(f"You can view in the UI here:\n   {self.ui_url()}")
        return True

    @cached_property
    def file_md5(self):
        with open(self.file_path, "rb") as f:
            return hash_io_hexdigest("md5", f)

    @property
    def file_basename(self):
        return os.path.basename(self.file_path)


class BugreportUploader(Uploader):
    authenticator_types = [
        ProjectKeyAuthenticator,
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        if not is_zipfile(self.file_path):
            LOG.error("%s is not a valid zip file!", self.file_path)
            return False
        return True

    def entity_url(self) -> str:
        if self.authenticator.project_key_auth():
            return f"{self._api_base_url()}/upload/bugreport"
        else:
            return f"{self._projects_base_url()}/bugreports"


class CoredumpUploader(Uploader):
    authenticator_types = [
        ProjectKeyAuthenticator,
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        with open(self.file_path, "rb") as f:
            hdr = f.read(4).decode("ascii", errors="ignore")
            if hdr != "CORE":
                LOG.error("%s is not a Memfault Coredump", self.file_path)
                return False

        return True

    def entity_url(self) -> str:
        if self.authenticator.project_key_auth():
            return f"{self._api_base_url()}/upload/coredump"
        else:
            return f"{self._projects_base_url()}/coredumps"

    def post_token_extra_request_data(self) -> Optional[Dict]:
        if self.ctx.core_info:
            return {
                "device_serial": self.ctx.core_info.device_serial,
            }
        return None


class XedUploader(Uploader):
    authenticator_types = [
        ProjectKeyAuthenticator,
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        return True

    def entity_url(self) -> str:
        if self.authenticator.project_key_auth():
            return f"{self._api_base_url()}/upload/qc/x"
        else:
            return f"{self._projects_base_url()}/qc/x"

    def post_token_extra_request_data(self) -> Optional[Dict]:
        info = self.ctx.xed_info
        return {
            "device_serial": info.device_serial,
            "hardware_version": info.hardware_version,
            "trace_attributes": [
                {"string_key": a.string_key, "value": a.value} for a in self.ctx.attributes
            ],
        }


class MarUploader(Uploader):
    authenticator_types = [
        ProjectKeyAuthenticator,
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        if not is_zipfile(self.file_path):
            LOG.error("%s is not a valid zip file!", self.file_path)
            return False
        return True

    def entity_url(self) -> str:
        if self.authenticator.project_key_auth():
            return f"{self._api_base_url()}/upload/mar"
        else:
            return f"{self._projects_base_url()}/mar"

    def post_token_extra_request_data(self) -> Optional[Dict]:
        info = self.ctx.mar_info
        return {
            "device_serial": info.device_serial,
            "hardware_version": info.hardware_version,
            "software_type": info.software_type,
            "software_version": info.software_version,
        }


class ElfSymbolUploader(Uploader):
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    gnu_build_id: str = ""

    def can_upload_file(self) -> bool:
        if os.path.islink(self.file_path) and not os.path.exists(self.file_path):
            return False

        with open(self.file_path, "rb") as f:
            if not is_elf(f):
                LOG.info("%s: Not an ELF file", self.file_path)
                return False

            elf = ELFFile(f)
            gnu_build_id, has_debug_info = get_gnu_build_id(elf), elf_has_debug_info(elf)
            if not gnu_build_id:
                LOG.info(
                    "%s: looks like an ELF but does not contain a GNU Build ID", self.file_path
                )
                return False

        self.gnu_build_id = gnu_build_id
        if not has_debug_info:
            LOG.info("%s: looks like an ELF but it has no .debug_info", self.file_path)
            return False
        LOG.info(
            "%s: ELF file with .debug_info and GNU Build ID: %s", self.file_path, self.gnu_build_id
        )
        return True

    def _is_already_uploaded(self) -> bool:
        response = self.session.head(
            f"{self._projects_base_url()}/symbols-by-gnu-build-id/{self.gnu_build_id}",
            **self.authenticator.requests_auth_params(),
        )
        try:
            check_response(response)
        except Exception:
            if response.status_code == 404:
                return False
            raise
        return True

    def entity_url(self) -> str:
        return f"{self._projects_base_url()}/symbols"


class McuSdkElfSymbolUploader(Uploader):
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        with open(self.file_path, "rb") as file:
            if file.read(4) != b"\x7FELF":
                LOG.info("%s: Not an ELF file", self.file_path)
                return False
            file.seek(0)

            inspector = BuildIdInspectorAndPatcher(file)
            if not elf_has_debug_info(inspector.elf):
                LOG.info("%s: looks like an ELF but it has no .debug_info", self.file_path)
                return False

            mcu_build_id_type, mcu_build_id, mcu_build_id_short_len = inspector.get_build_info()

            # NB: QC symbol files have different arg requirements depending on the subsystem the
            # ELF is associated with so we offload to the Memfault backend.
            if not self.ctx.software_info and mcu_build_id is None and not self.qcomm_project:
                LOG.error(
                    "%s: Build Id missing. Specify --software-version and --software-type options "
                    "or add a Build Id (see https://mflt.io/symbol-file-build-ids)",
                    self.file_path,
                )
                return False

        LOG.info("%s: ELF file with .debug_info. Build Id: %s", self.file_path, mcu_build_id)
        return True

    def _is_already_uploaded(self) -> bool:
        return False

    def entity_url(self) -> str:
        return f"{self._projects_base_url()}/symbols"

    def ui_url(self) -> Optional[str]:
        if not self.ctx.software_info or (
            # Unreachable with our CLI, but for safety:
            isinstance(self.ctx.software_info.version, DeltaReleaseVersion)
        ):
            return None

        software_type = url_quote(self.ctx.software_info.software_type)
        software_version = url_quote(self.ctx.software_info.version.software_version)
        return (
            f"{self._projects_ui_base_url()}/software/{software_type}/versions/{software_version}"
        )

    def post_token_extra_request_data(self) -> Optional[Dict]:
        if not self.ctx.software_info:
            return None

        info = self.ctx.software_info
        return {
            "software_version": {
                "version": info.version.to_json_representable(),
                "software_type": info.software_type,
                **({"revision": info.revision} if info.revision else {}),
            },
        }


class ReleaseArtifactUploader(Uploader):
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        # There are no restrictions for the type of binary a customer uses for
        # a release
        return True

    def _is_already_uploaded(self) -> bool:
        return False

    def entity_url(self) -> str:
        return f"{self._projects_base_url()}/releases/ota_payload"

    def ui_url(self) -> str:
        info = self.ctx.software_info
        releases_url = f"{self._projects_ui_base_url()}/releases"
        delta_releases_url = f"{self._projects_ui_base_url()}/delta-releases"

        if not info:
            return releases_url

        if info.software_type == "android-build":
            # In case of Android, the created Release.version may not be the same
            # as the SoftwareVersion.version, in case build fingerprints are used,
            # the Release.version will be "stripped" (i.e. */*/*:*/*/*:*/*::1.0.0):
            return (
                delta_releases_url
                if isinstance(info.version, DeltaReleaseVersion)
                else releases_url
            )

        if isinstance(info.version, DeltaReleaseVersion):
            from_version = url_quote(info.version.delta_from)
            to_version = url_quote(info.version.delta_to)
            return f"{delta_releases_url}/{from_version}/{to_version}"

        return f"{releases_url}/{url_quote(info.version.software_version)}"

    def post_token_extra_request_data(self) -> Optional[Dict]:
        info = self.ctx.software_info
        assert info  # mypy

        return {
            "software_version": {
                "version": info.version.to_json_representable(),
                "software_type": info.software_type,
                **({"revision": info.revision} if info.revision else {}),
            },
            "must_pass_through": self.ctx.must_pass_through,
            "hardware_version": self.ctx.hardware_version,
            "notes": self.ctx.notes,
        }


def walk_files(toplevel: str) -> Iterable[str]:
    for root, _dirs, files in os.walk(toplevel):
        for file in files:
            yield os.path.join(root, file)


class ParallelUploader(Uploader):
    uploader_cls: Type[Uploader]

    def __init__(
        self,
        *args,
        file_paths: Iterable[Union[str, os.PathLike]],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.file_paths = file_paths

    def can_upload_file(self) -> bool:
        if os.path.islink(self.file_path) and not os.path.exists(self.file_path):
            return False

        return os.path.isdir(self.file_path)

    def upload(self, *, progressbar=True) -> bool:
        queued_uploaders = []
        did_upload = False

        for file_path in self.file_paths:
            uploader = self.uploader_cls(
                ctx=self.ctx, file_path=file_path, authenticator=self.authenticator
            )
            if uploader.can_upload_file():
                queued_uploaders.append(uploader)
                did_upload |= True
            else:
                LOG.info("%s: skipping...", file_path)

        def do_upload(uploader: Uploader):
            uploader.upload(progressbar=False)

        # If we find that we have more implementations that need to do parallel uploads,
        # We should generalize this logic into a class or have a function called `get_uploaders`
        # that the calling client can then parallelize by calling `.upload()` on each.
        with ThreadPoolExecutor(max_workers=self.ctx.concurrency) as pool:
            # NB: We need to read the result for exceptions to be raised
            for result in pool.map(do_upload, queued_uploaders):
                del result

        return did_upload

    def entity_url(self) -> str:
        raise NotImplementedError


class ElfSymbolDirectoryUploader(ParallelUploader):
    uploader_cls = ElfSymbolUploader
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]


class ProguardMappingUploader(Uploader):
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def __init__(
        self,
        *,
        ctx: MemfaultCliClickContext,
        file_path: str,
        authenticator: Authenticator,
        package_info: PackageInfo,
        **kwargs,
    ):
        super().__init__(ctx=ctx, file_path=file_path, authenticator=authenticator, **kwargs)
        self.package_info = package_info

    def can_upload_file(self) -> bool:
        return True

    def entity_url(self) -> str:
        return f"{self._projects_base_url()}/symbols"

    def post_token_extra_request_data(self) -> Optional[Dict]:
        return {
            "software_version": {
                "version": f"{self.package_info.version_name}:{self.package_info.version_code}",
                "software_type": self.package_info.name,
            },
        }


class AndroidAppSymbolsUploader(Uploader):
    authenticator_types = [
        OrgTokenAuthenticator,
        BasicAuthenticator,
    ]

    def can_upload_file(self) -> bool:
        if not os.path.isdir(self.file_path):
            LOG.error(
                '%s is not a directory. Please specify the Android app\'s root "build" path instead!',
                self.file_path,
            )
            return False
        return True

    def _find_apk_path(self) -> str:
        apk_dir = os.path.join(self.file_path, "outputs", "apk", self.ctx.build_variant)
        if not os.path.isdir(apk_dir):
            raise click.exceptions.FileError(
                apk_dir, hint='Please pass the Android app\'s root "build" path!'
            )
        apk_glob = os.path.join(apk_dir, "*.apk")
        apks = glob(apk_glob)
        if not apks:
            raise click.exceptions.FileError(apk_glob, hint="No .apk found!")
        if len(apks) > 1:
            raise click.exceptions.FileError(apk_glob, hint="Multiple .apks found!")
        return apks[0]

    def _get_package_info_from_apk(self) -> PackageInfo:
        from pyaxmlparser import APK

        apk = APK(self._find_apk_path())
        return PackageInfo(apk.get_package(), apk.version_name, apk.version_code)

    def _find_mapping_txt(self) -> Optional[str]:
        if self.ctx.android_mapping_txt:
            return self.ctx.android_mapping_txt
        # In older version, the default folder was called "proguard", later it got renamed to "mapping":
        subdirs = ("mapping", "proguard")
        paths_to_check = tuple(
            map(
                lambda subdir: os.path.join(
                    self.file_path, "outputs", "mapping", self.ctx.build_variant, "mapping.txt"
                ),
                subdirs,
            )
        )
        for path in paths_to_check:
            if os.path.isfile(path):
                return path
        LOG.info("No mapping.txt found in default locations, assuming ProGuard/R8 is disabled...")
        return None

    def _upload_mapping_txt(self) -> bool:
        mapping_txt_path = self._find_mapping_txt()
        if not mapping_txt_path:
            return False
        package_info = self.ctx.android_package_info
        if not package_info:
            package_info = self._get_package_info_from_apk()
        pg_uploader = ProguardMappingUploader(
            ctx=self.ctx,
            file_path=mapping_txt_path,
            package_info=package_info,
            authenticator=self.authenticator,
        )
        return pg_uploader.upload()

    def _upload_ndk_symbols(self) -> bool:
        # TODO: add support for legacy ndkBuild?
        cmake_path_out = os.path.join(
            self.file_path, "intermediates", "cmake", self.ctx.build_variant
        )
        if not os.path.isdir(cmake_path_out):
            return False
        sym_dir_uploader = ElfSymbolDirectoryUploader(
            ctx=self.ctx,
            file_path=cmake_path_out,
            authenticator=self.authenticator,
            file_paths=walk_files(cmake_path_out),
        )
        return sym_dir_uploader.upload()

    def upload(self, *, progressbar=True) -> bool:
        did_upload = self._upload_mapping_txt() or self._upload_ndk_symbols()
        if not did_upload:
            raise click.exceptions.UsageError("No files uploaded!")
        return did_upload

    def entity_url(self) -> str:
        raise NotImplementedError



import logging
from typing import Dict, Tuple, Union

import requests
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from .authenticator import Authenticator
from .context import MemfaultCliClickContext

LOG = logging.getLogger(__name__)


class Deployer:
    def __init__(self, *, ctx: MemfaultCliClickContext, authenticator: Authenticator):
        self.ctx: MemfaultCliClickContext = ctx
        self.authenticator: Authenticator = authenticator
        self.session = self._create_requests_session()

    @staticmethod
    def _create_requests_session() -> Session:
        retry_strategy = Retry(
            total=5,
            backoff_factor=2,  # Sleep for 2s, 4s, 8s, 16s, 32s, Stop
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "POST", "PUT"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        return session

    @property
    def base_url(self) -> str:
        return self.ctx.api_url

    def _api_base_url(self) -> str:
        return f"{self.base_url}/api/v0"

    def _projects_base_url(self) -> str:
        return f"{self.base_url}/api/v0/organizations/{self.ctx.org}/projects/{self.ctx.project}"

    @property
    def deployment_url(self) -> str:
        """
        The upload URL for a 'prepared upload' for the configured authentication method.
        """
        return f"{self._projects_base_url()}/deployments"

    def deploy(
        self, *, release_version: Union[str, Tuple[str, str]], cohort: str, rollout_percent: int
    ) -> None:
        deployment_type = "normal" if rollout_percent == 100 else "staged_rollout"
        json_d: Dict[str, Union[str, Tuple[str, str], int]] = {
            "type": deployment_type,
            "release": release_version,
            "cohort": cohort,
        }

        if deployment_type == "staged_rollout":
            json_d["rollout_percent"] = rollout_percent

        response = self.session.post(
            self.deployment_url, json=json_d, **self.authenticator.requests_auth_params()
        )
        if response.status_code >= 400:
            raise Exception(
                f"Request failed with HTTP status {response.status_code}\nResponse body:\n{response.content.decode()}"
            )
        LOG.info("Release %s successfully deployed to Cohort %s", release_version, cohort)



import dataclasses
import logging
import pathlib
from typing import Dict, Iterable, TextIO, Tuple

import more_itertools

from .elf import find_elf_files

# See https://docs.yoctoproject.org/overview-manual/concepts.html#package-splitting
# for the directory structure of the Yocto build output.

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class ManifestEntry:
    package: str  # Note this is the name *after* package splitting!
    arch: str
    version: str

    @staticmethod
    def from_line(line: str) -> "ManifestEntry":
        components = line.split(maxsplit=3)
        assert len(components) == 3, "Manifest line has unexpected number of items"
        return ManifestEntry(*components)


def parse_image_manifest(file: Iterable[str]) -> Iterable[ManifestEntry]:
    for line in file:
        yield ManifestEntry.from_line(line)


def tmp_path_from_manifest(manifest: TextIO) -> pathlib.Path:
    """
    This function returns the resolved tmp/ given a .manifest file's path.
    The .manifest file is created by Yocto in tmp/deploy/images/<arch>/.
    """
    manifest_path = pathlib.Path(manifest.name).resolve()
    return manifest_path.parent.parent.parent.parent


def gather_workdirs_by_split_package_names_and_version(
    arch_path: pathlib.Path,
) -> Dict[Tuple[str, str], pathlib.Path]:
    """
    The manifest file holds package names *after* splitting. We need to map these back to the paths where each was
    built ("image" folder). The folder for each package before splitting has a "packages-split" folder. In this
    folder, there are subfolders named after the packages *after* splitting.
    The directory structure in the arch path looks like this:
    - cortexa57-poky-linux  <- arch_path
        - PN (package name before splitting)
            - PV-PR (version - revision) <- package's WORKDIR
                - image <- folder containing "raw" .elfs (symbols)
                - packages-split
                    - PN-1 (packages name after splitting()
                    - PN-2
                    - PN..
    :param arch_path: The arch path to search in.
    :return: Dict mapping (package name after splitting, package version) => package's WORKDIR
    """
    mapping = {}

    for package_dir in arch_path.iterdir():
        for package_workdir in package_dir.iterdir():
            packages_split_dir = package_workdir / "packages-split"
            if not packages_split_dir.exists():
                mapping[(package_dir.name, package_workdir.name)] = package_workdir
                continue
            for dir in packages_split_dir.iterdir():
                mapping[(dir.name, package_workdir.name)] = package_workdir

    return mapping


def find_package_workdirs_from_image_manifest_entries(
    entries: Iterable[ManifestEntry], tmp_path: pathlib.Path
) -> Iterable[pathlib.Path]:
    """
    This function finds package folders matching the given manifest file.
    :return: Iterable with found package WORKDIRs.
    """
    entries_by_arch = more_itertools.map_reduce(entries, lambda e: e.arch)
    work_path = tmp_path / "work"
    if not work_path.exists():
        raise FileNotFoundError(
            f"Missing {work_path}. The .manifest file is expected to "
            f"be in the location where it was originally put by Yocto."
        )
    for arch_path in work_path.iterdir():
        if not arch_path.is_dir():
            continue
        workdir_by_package_name_and_version = gather_workdirs_by_split_package_names_and_version(
            arch_path
        )
        arch, _, _ = arch_path.name.partition("-")
        entries_for_arch = entries_by_arch.get(arch)
        if not entries_for_arch:
            LOG.debug("No manifest entries from arch: %s", arch)
            continue
        for entry in entries_for_arch:
            package_workdir = workdir_by_package_name_and_version.get(
                (entry.package, entry.version)
            )
            if package_workdir:
                yield package_workdir
            else:
                LOG.debug("Could not find package workdir for %s", entry)


def should_ignore_elf(rel_elf_path: pathlib.Path) -> bool:
    # Ignore ptests -- See https://wiki.yoctoproject.org/wiki/Ptest
    if rel_elf_path.suffix in {".a", ".o"}:
        LOG.debug("Ignoring %s (static lib / object file)", rel_elf_path)
        return True
    if any((parent.name == "ptest" for parent in rel_elf_path.parents)):
        LOG.debug("Ignoring %s (ptest)", rel_elf_path)
        return True
    return False


def should_ignore_package(entry: ManifestEntry) -> bool:
    _, _, ending = entry.package.rpartition("-")
    if ending.startswith("locale-"):
        # Ignore all *-locale-* packages (optimization).
        return True
    return ending in {
        # Ignored because they contain split .elf files (only .debug_info and no .text/.data):
        "dbg",
        # Ignored because they contain static libraries, not executables:
        "staticdev",
        # Ignored because they contain test executables:
        "ptest",
        # Ignore source, doc and locale packages (optimization):
        "src",
        "doc",
        "locale",
    }


def find_elf_files_from_image_manifest(manifest: TextIO) -> Iterable[pathlib.Path]:
    entries = (e for e in parse_image_manifest(manifest) if not should_ignore_package(e))
    tmp_path = tmp_path_from_manifest(manifest)
    for workdir in more_itertools.unique_everseen(
        find_package_workdirs_from_image_manifest_entries(entries, tmp_path)
    ):
        for elf_file in find_elf_files(workdir / "image"):
            if should_ignore_elf(elf_file.relative_to(workdir)):
                continue
            yield elf_file



import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from .context import MemfaultCliClickContext


LOG = logging.getLogger(__name__)


class Authenticator:
    def __init__(self, ctx: "MemfaultCliClickContext"):
        self.ctx = ctx

    @staticmethod
    @abstractmethod
    def project_key_auth() -> bool:
        pass

    @classmethod
    @abstractmethod
    def required_args(cls) -> List[str]:
        pass

    @abstractmethod
    def requests_auth_params(self) -> dict:
        pass


class ProjectKeyAuthenticator(Authenticator):
    """
    Project Key Authentication with the Memfault service (Memfault-Project-Key)
    """

    @staticmethod
    def project_key_auth() -> bool:
        return True

    @classmethod
    def required_args(cls) -> List[str]:
        return [
            "project_key",
        ]

    def requests_auth_params(self) -> dict:
        return dict(headers={"Memfault-Project-Key": self.ctx.project_key})


class BasicAuthenticator(Authenticator):
    """
    Basic Authentication with the Memfault service

    Various ways a user can authenticate:
    - Email + Password
    - Email + User API Key
    - DEPRECATED: Organization Auth Token (passed in as password) -- --org-token should be used instead
    """

    def __init__(self, ctx: "MemfaultCliClickContext") -> None:
        super().__init__(ctx)
        if not self.ctx.obj.get("email") and self.ctx.password.startswith("oat_"):
            LOG.warning(
                "Please use --org-token instead of --password to pass organization auth token"
            )

    @staticmethod
    def project_key_auth() -> bool:
        return False

    @classmethod
    def required_args(cls) -> List[str]:
        return [
            "org",
            "project",
            "password",
        ]

    def requests_auth_params(self) -> dict:
        # Email can be None in the case of using Organization auth tokens (prefixed with oat_*)
        auth = (self.ctx.obj.get("email") or "", self.ctx.password)
        return dict(auth=auth)


class OrgTokenAuthenticator(Authenticator):
    """
    Org Token Authentication with the Memfault service, passed in with --org-token
    """

    @staticmethod
    def project_key_auth() -> bool:
        return False

    @classmethod
    def required_args(cls) -> List[str]:
        return [
            "org",
            "org_token",
            "project",
        ]

    def requests_auth_params(self) -> dict:
        auth = ("", self.ctx.org_token)
        return dict(auth=auth)



import json
import logging
from typing import Any, Dict, List, Optional, Set, Type, Union

import click

from .authenticator import Authenticator

LOG = logging.getLogger(__name__)


class PackageInfo:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, name: str, version_name: str, version_code: str):
        self.name = name
        self.version_name = version_name
        self.version_code = version_code


class FullReleaseVersion:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, software_version: str):
        self.software_version = software_version

    def to_json_representable(self):
        return self.software_version


class DeltaReleaseVersion:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, delta_from: str, delta_to: str):
        self.delta_from = delta_from
        self.delta_to = delta_to

    def to_json_representable(self):
        return [self.delta_from, self.delta_to]


class SoftwareInfo:  # noqa: B903 - no data classes in Python 3.6
    def __init__(
        self,
        software_type: str,
        version: Union[FullReleaseVersion, DeltaReleaseVersion],
        revision: Optional[str] = None,
    ):
        self.software_type = software_type
        self.version = version
        self.revision = revision


CustomMetricReadingValueType = Union[int, float, str, bool, None]


class Attribute:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, string_key: str, value: CustomMetricReadingValueType) -> None:
        self.string_key = string_key
        self.value = value

    @staticmethod
    def from_cli_args(key: str, cli_value: str):
        return Attribute(string_key=key, value=Attribute._convert_value(cli_value))

    @staticmethod
    def _convert_value(cli_value: str) -> CustomMetricReadingValueType:
        if cli_value == "":
            return None
        try:
            value = json.loads(cli_value)
        except json.decoder.JSONDecodeError:
            value = cli_value
        if value is not None and not isinstance(value, (int, float, str, bool)):
            LOG.warning(
                "Only boolean, number, string or null types are allowed! Treating passed value as a string."
            )
            return cli_value
        return value


class XedInfo:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, device_serial: str, hardware_version: str):
        self.device_serial = device_serial
        self.hardware_version = hardware_version


class CoreInfo:  # noqa: B903 - no data classes in Python 3.6
    def __init__(self, device_serial: str):
        self.device_serial = device_serial


class MarInfo:  # noqa: B903 - no data classes in Python 3.6
    def __init__(
        self, device_serial: str, hardware_version: str, software_type: str, software_version: str
    ):
        self.device_serial = device_serial
        self.hardware_version = hardware_version
        self.software_type = software_type
        self.software_version = software_version


class MemfaultCliClickContext(object):
    """
    A context passed around between the memfault cli sub-commands.


    If the top level CLI has any "required" it's not possible to display
    any help info about the subcommands using "--help" without providing them.
    By passing around this context, subcommand help messages can be displayed
    and errors can be raised in a uniform way
    """

    def __init__(self) -> None:
        self.obj: Dict[str, Any] = {}

    def _format_as_option(self, property_name: str) -> str:
        return f"--{property_name.replace('_', '-')}"

    def _find_obj_or_raise(self, name):
        value = self.obj.get(name)
        if value is None:
            raise click.exceptions.UsageError(f"Missing option {self._format_as_option(name)!r}.")
        return value

    def check_required_auth_cli_args(self, *, authenticator_type: Type[Authenticator]):
        required_args = authenticator_type.required_args()

        for arg in required_args:
            self._find_obj_or_raise(arg)

    def _format_arg_set(self, arg_set: Set[str]) -> str:
        return ", ".join([f"{self._format_as_option(arg)!r}" for arg in sorted(arg_set)])

    def check_required_either(self, *args: Set[str], mutually_exclusive=False):
        """Check that any full set of args is present"""
        valid_arg_sets = []
        passed_arg_sets = []

        for arg_set in args:
            args_present = [self.obj.get(arg, None) is not None for arg in arg_set]
            if any(args_present):
                passed_arg_sets.append(arg_set)
            if all(args_present):
                valid_arg_sets.append(arg_set)

        if mutually_exclusive and len(passed_arg_sets) > 1:
            formatted_passed_args = [self._format_arg_set(arg_set) for arg_set in passed_arg_sets]
            raise click.exceptions.UsageError(
                f"Parameters {' and '.join(sorted(formatted_passed_args))} may not be used together."
            )

        if len(valid_arg_sets) == 0:
            formatted_required_args = [self._format_arg_set(arg_set) for arg_set in args]
            raise click.exceptions.UsageError(
                f"Please pass either {' or '.join(sorted(formatted_required_args))}."
            )

    def create_authenticator(self, *types: Type[Authenticator]) -> Authenticator:
        assert types
        for _type in types:
            try:
                self.check_required_auth_cli_args(authenticator_type=_type)
                return _type(self)
            except click.exceptions.UsageError:
                pass
        # Raise for the first missing option of the preferred authenticator type:
        raise self.check_required_auth_cli_args(authenticator_type=types[0])

    @property
    def org(self):
        return self._find_obj_or_raise("org")

    @property
    def org_token(self) -> str:
        return self._find_obj_or_raise("org_token")

    @property
    def project(self):
        return self._find_obj_or_raise("project")

    @property
    def email(self):
        return self._find_obj_or_raise("email")

    @property
    def password(self):
        return self._find_obj_or_raise("password")

    @property
    def project_key(self):
        return self._find_obj_or_raise("project_key")

    @property
    def concurrency(self):
        return self._find_obj_or_raise("concurrency")

    @property
    def software_info(self) -> Optional[SoftwareInfo]:
        sw_type = self.obj.get("software_type")
        sw_ver = self.obj.get("software_version")
        revision = self.obj.get("revision")
        delta_from = self.obj.get("delta_from")
        delta_to = self.obj.get("delta_to")

        version: Union[DeltaReleaseVersion, FullReleaseVersion, None]
        if delta_from and delta_to:
            version = DeltaReleaseVersion(delta_from, delta_to)
            if delta_from == delta_to:
                raise click.exceptions.UsageError(
                    "Version passed to --delta-{{from-to}} must not be the same, please specify two different versions"
                )
        elif sw_ver:
            version = FullReleaseVersion(sw_ver)
        else:
            version = None

        if revision and (version is None or sw_type is None):
            raise click.exceptions.UsageError(
                "A version ('--software-version' or '--delta-{{from,to}}') and '--software-type' must be specified when using '--revision'"
            )

        if sw_type is None and version is None:
            return None

        if sw_type is None or version is None:
            raise click.exceptions.UsageError(
                "Version ('--software-version' or '--delta-{{from,to}}') and '--software-type' must be specified together"
            )

        return SoftwareInfo(
            software_type=sw_type,
            version=version,
            revision=revision,
        )

    @property
    def attributes(self) -> List[Attribute]:
        return [Attribute.from_cli_args(*cli_args) for cli_args in self.obj.get("attribute", [])]

    @property
    def core_info(self) -> Optional[CoreInfo]:
        device_serial = self.obj.get("device_serial")
        if device_serial:
            return CoreInfo(
                device_serial=device_serial,
            )
        return None

    @property
    def xed_info(self) -> XedInfo:
        device_serial = self._find_obj_or_raise("device_serial")
        hardware_version = self._find_obj_or_raise("hardware_version")
        return XedInfo(
            device_serial=device_serial,
            hardware_version=hardware_version,
        )

    @property
    def mar_info(self) -> MarInfo:
        device_serial = self._find_obj_or_raise("device_serial")
        hardware_version = self._find_obj_or_raise("hardware_version")
        software_type = self._find_obj_or_raise("software_type")
        software_version = self._find_obj_or_raise("software_version")
        return MarInfo(
            device_serial=device_serial,
            hardware_version=hardware_version,
            software_type=software_type,
            software_version=software_version,
        )

    @property
    def file_url(self) -> str:
        url = self.obj.get("url")
        if url is None:
            return "https://files.memfault.com"
        return url

    @property
    def app_url(self) -> str:
        url = self.obj.get("url")
        if url is None:
            return "https://app.memfault.com"
        return url

    @property
    def chunks_url(self) -> str:
        url = self.obj.get("url")
        if url is None:
            return "https://chunks.memfault.com"
        return url

    @property
    def api_url(self) -> str:
        url = self.obj.get("url")
        if url is None:
            return "https://api.memfault.com"
        return url

    @property
    def hardware_version(self):
        return self.obj.get("hardware_version", None)

    @property
    def device_serial(self):
        return self._find_obj_or_raise("device_serial")

    @property
    def build_variant(self):
        return self._find_obj_or_raise("build_variant")

    @property
    def must_pass_through(self):
        return self._find_obj_or_raise("must_pass_through")

    @property
    def notes(self):
        # release notes for an OTA package
        return self._find_obj_or_raise("notes")

    @property
    def android_package_info(self) -> Optional[PackageInfo]:
        package = self.obj.get("package")
        version_name = self.obj.get("version_name")
        version_code = self.obj.get("version_code")
        if package is None and version_name is None and version_code is None:
            return None
        if package is None or version_name is None or version_code is None:
            raise click.exceptions.UsageError(
                '"--package, --version-name" and "--version-code" must be specified together'
            )

        return PackageInfo(package, version_name, version_code)

    @property
    def android_mapping_txt(self) -> Optional[str]:
        return self.obj.get("mapping_txt")

    @property
    def verbose(self) -> bool:
        return self.obj.get("verbose", False)



try:
    from importlib.metadata import version as get_version
    from importlib.metadata import PackageNotFoundError
except ImportError:
    # fallback for python <3.8
    from importlib_metadata import PackageNotFoundError  # type: ignore[misc]

# when running tests on the repo, provide a fallback value, since the
# memfault-cli package is not installed at that time
try:
    version = get_version(__package__)
except PackageNotFoundError:
    version = "dev"



import pathlib
from typing import BinaryIO, Iterable, Optional

import more_itertools
from elftools.construct import Container
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import NoteSection
from elftools.elf.segments import NoteSegment

# FIXME: Deduplicate. Copy pasta from elf_utils.py

NT_GNU_BUILD_ID = "NT_GNU_BUILD_ID"

ELF_NOTE_SECTION_OWNER_GNU = "GNU"


def is_elf(file: BinaryIO) -> bool:
    original_offset = file.tell()
    magic = file.read(4)
    file.seek(original_offset)
    return magic == b"\x7FELF"


def elf_has_debug_info(elf: ELFFile) -> bool:
    # Note: not using .has_dwarf_info() because it will return True if .eh_frame is present.
    # This section is usually kept in stripped binaries, because certain languages (C++) depend on it at runtime.
    return bool(elf.get_section_by_name(".debug_info") or elf.get_section_by_name(".zdebug_info"))


def find_elf_files(dir: pathlib.Path, *, recurse: bool = True) -> Iterable[pathlib.Path]:
    if not dir.is_dir() or not dir.exists():
        return
    for path in dir.iterdir():
        if path.is_dir() and recurse:
            yield from find_elf_files(path)
        elif path.is_file():
            with open(path, "rb") as f:
                if not is_elf(f):
                    continue
            yield path


def get_note_segments(elf: ELFFile) -> Iterable[NoteSegment]:
    return filter(lambda segment: isinstance(segment, NoteSegment), elf.iter_segments())


def get_note_sections(elf: ELFFile) -> Iterable[NoteSection]:
    return filter(lambda segment: isinstance(segment, NoteSection), elf.iter_sections())


def get_notes(elf: ELFFile) -> Iterable[Container]:
    for note_segment in get_note_segments(elf):
        yield from note_segment.iter_notes()
    for note_section in get_note_sections(elf):
        yield from note_section.iter_notes()


def is_gnu_build_id_note_section(section: NoteSection) -> bool:
    return (section.n_type == NT_GNU_BUILD_ID) and (section.n_name == ELF_NOTE_SECTION_OWNER_GNU)


def get_gnu_build_id(elf: ELFFile) -> Optional[str]:
    build_id_note = more_itertools.first_true(
        get_notes(elf),
        pred=is_gnu_build_id_note_section,
    )
    if not build_id_note:
        return None
    return build_id_note.n_desc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['memfault_cli']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7,<9',
 'mflt-build-id==0.0.5',
 'more_itertools>=8.0.2,<9.0.0',
 'pyaxmlparser>=0.3.24,<0.4.0',
 'pyelftools>=0.26,<=0.28',
 'requests>=2.22.0,<3.0.0',
 'tqdm>=4.44.1,<5.0.0',
 'urllib3>=1.26.7']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata==4.8.3']}

entry_points = \
{'console_scripts': ['memfault = memfault_cli.cli:main']}

setup_kwargs = {
    'name': 'memfault-cli',
    'version': '0.9.0',
    'description': 'Memfault CLI tool',
    'long_description': '# Memfault CLI tool\n\nThis package contains the `memfault` CLI tool.\n\nThe purpose of the tool is to make integration with Memfault from other systems,\nlike continuous integration servers, as easy as possible.\n\nInstall the tool and run `memfault --help` for more info!\n',
    'author': 'Memfault Inc',
    'author_email': 'hello@memfault.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://docs.memfault.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import http.client
import logging
import os
from base64 import b64decode
from typing import List, TextIO, Type

import click

from ._version import version
from .authenticator import BasicAuthenticator, OrgTokenAuthenticator, ProjectKeyAuthenticator
from .chunk import MemfaultChunk
from .context import MemfaultCliClickContext
from .deploy import Deployer
from .upload import (
    AndroidAppSymbolsUploader,
    BugreportUploader,
    CoredumpUploader,
    ElfSymbolDirectoryUploader,
    ElfSymbolUploader,
    MarUploader,
    McuSdkElfSymbolUploader,
    ReleaseArtifactUploader,
    Uploader,
    XedUploader,
    walk_files,
)
from .yocto import find_elf_files_from_image_manifest

pass_memfault_cli_context = click.make_pass_decorator(MemfaultCliClickContext, ensure=True)
click_argument_path = click.argument("path", type=click.Path(exists=True))
click_option_concurrency = click.option(
    "--concurrency",
    required=False,
    default=8,
    type=int,
    help="Max number of concurrent web requests",
)
click_option_revision = click.option(
    "--revision", help="Revision SHA or # (git, SVN, etc.)", required=False
)


@click.group()
@click.option("--email", help="Account email to authenticate with")
@click.password_option(
    "--password", prompt=False, help="Account password or user API key to authenticate with"
)
@click.option("--project-key", help="Memfault Project Key")
@click.option("--org-token", help="Organization Auth Token")
@click.option("--org", help="Organization slug")
@click.option("--project", help="Project slug")
@click.option("--url", hidden=True)
@click.option("--verbose", help="Log verbosely", is_flag=True)
@click.version_option(version=version)
@pass_memfault_cli_context
def cli(ctx: MemfaultCliClickContext, **kwargs):
    ctx.obj.update(kwargs)

    if ctx.verbose:
        click.echo(f"version: {version}")

        logging.basicConfig(level=logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

        httpclient_logger = logging.getLogger("http.client")
        httpclient_logger.setLevel(logging.DEBUG)

        def httpclient_log(*args) -> None:
            httpclient_logger.log(logging.DEBUG, " ".join(args))

        http.client.print = httpclient_log  # type: ignore[attr-defined]
        http.client.HTTPConnection.debuglevel = 1


def _do_upload_or_raise(
    ctx: MemfaultCliClickContext, path: str, uploader_cls: Type[Uploader], **kwargs
) -> None:
    authenticator = ctx.create_authenticator(*uploader_cls.authenticator_types)

    uploader = uploader_cls(ctx=ctx, file_path=path, authenticator=authenticator, **kwargs)
    if not uploader.can_upload_file():
        raise click.exceptions.UsageError("Upload failed!")
    uploader.upload()


@cli.command(name="upload-coredump")
@click.option(
    "--device-serial",
    required=False,
    help="A fallback unique identifier of a device, when one is not present in the coredump",
)
@click_argument_path
@pass_memfault_cli_context
def upload_coredump(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload an MCU coredump for analysis.

    Coredumps can be added to an MCU platform by integrating the Memfault C SDK:
    https://github.com/memfault/memfault-firmware-sdk
    """
    ctx.obj.update(**kwargs)

    _do_upload_or_raise(ctx, path, CoredumpUploader)


@cli.command(name="upload-xed")
@click.option(
    "--hardware-version",
    required=True,
    help="Required for MCU builds, see https://mflt.io/34PyNGQ",
)
@click.option(
    "--device-serial",
    required=True,
    help="The unique identifier of a device",
)
@click.option(
    "--attribute",
    type=click.Tuple([str, str]),
    multiple=True,
    help="""
    Attribute associated with the uploaded file. Multiple attributes can be passed.
    The value argument is attempted to be interpreted as a JSON string first, if that
    fails, the value is interpreted as a string value as-is. Note however that only
    number, boolean, string or null values are allowed (objects and arrays are not allowed).

    Examples:

    Key "my_int" with numeric value 1234:

    --attribute my_int 1234

    Key "my_string" with string value 'hello':

    --attribute my_string hello

    Key "my_bool" with boolean true value:

    --attribute my_bool true
    """,
)
@click_argument_path
@pass_memfault_cli_context
def upload_xed(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload an .xed or .xcd file for analysis."""
    ctx.obj.update(**kwargs)

    _do_upload_or_raise(ctx, path, XedUploader)


@cli.command(name="upload-bugreport")
@click_argument_path
@pass_memfault_cli_context
def upload_bugreport(ctx: MemfaultCliClickContext, path: str):
    """Upload an Android Bug Report for analysis by Memfault."""
    uploader_cls = BugreportUploader
    _do_upload_or_raise(ctx, path, uploader_cls)


@cli.command(name="upload-symbols")
@click.option(
    "--software-type",
    required=False,
    help="Required for MCU builds, see https://mflt.io/34PyNGQ",
)
@click.option(
    "--software-version",
    required=False,
    help="Required for MCU builds, see https://mflt.io/34PyNGQ",
)
@click_option_revision
@click_option_concurrency
@click_argument_path
@pass_memfault_cli_context
def upload_symbols(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """[DEPRECATED] Upload symbols for an MCU or Android build.

    Please use upload-aosp-symbols or upload-mcu-symbols instead!
    """
    ctx.obj.update(**kwargs)

    uploader: Type[Uploader]
    if os.path.isdir(path):
        uploader = ElfSymbolDirectoryUploader
        uploader_kwargs = dict(file_paths=walk_files(path))
    elif ctx.software_info is None:
        uploader = ElfSymbolUploader
        uploader_kwargs = dict()
    else:
        uploader = McuSdkElfSymbolUploader
        uploader_kwargs = dict()

    _do_upload_or_raise(ctx, path, uploader, **uploader_kwargs)


@cli.command(name="upload-mar")
@click.option(
    "--hardware-version",
    required=True,
    help="Required to identify the type of Android hardware, see https://mflt.io/34PyNGQ/#android. By default the ro.product.board property is used.",
)
@click.option(
    "--software-type",
    required=True,
    help="Required to identify the Android system software, see https://mflt.io/34PyNGQ/#android-1",
)
@click.option(
    "--software-version",
    required=True,
    help="Required to identify single builds on Android devices, see https://mflt.io/34PyNGQ/#software-version",
)
@click.option(
    "--device-serial",
    required=True,
    help="The unique identifier of a device",
)
@click_argument_path
@pass_memfault_cli_context
def upload_mar(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload a Memfault Archive File (mar) file for analysis."""
    ctx.obj.update(**kwargs)
    _do_upload_or_raise(ctx, path, MarUploader)


@cli.command(name="upload-yocto-symbols")
@click.option(
    "--manifest",
    "-m",
    required=True,
    help="""
    The path to the image .manifest as produced by Yocto's do_rootfs (IMAGE_MANIFEST).
    The .manifest file is expected to reside in the location where Yocto produced the file.
    The tool will locate the symbol files to upload in the tmp/work/ directory, based on
    the packages that are listed in the .manifest file.
    """,
    type=click.File(),
)
@click_option_concurrency
@pass_memfault_cli_context
def upload_yocto_symbols(ctx: MemfaultCliClickContext, manifest: TextIO, **kwargs):
    """Upload symbols for a Yocto build.

    To see a full example, take a look at the Linux SDK example project:
    https://mflt.io/yocto-symbols-image

    Example Yocto Symbol Upload:

        \b
        $ bitbake my-image
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
            upload-yocto-symbols --manifest build/tmp/deploy/images/raspberrypi3/my-image-raspberrypi3.rootfs.manifest
    """
    ctx.obj.update(**kwargs)

    ElfSymbolDirectoryUploader(
        ctx=ctx,
        file_path=manifest.name,  # Not used. TODO: refactor!
        authenticator=ctx.create_authenticator(OrgTokenAuthenticator, BasicAuthenticator),
        file_paths=find_elf_files_from_image_manifest(manifest),
    ).upload()


@cli.command(name="upload-mcu-symbols")
@click.option(
    "--software-type",
    required=False,
    help="Required for MCU symbols without Build Id, see https://mflt.io/symbol-file-build-ids",
)
@click.option(
    "--software-version",
    required=False,
    help="Required for MCU symbols without Build Id, see https://mflt.io/symbol-file-build-ids",
)
@click_option_revision
@click_argument_path
@pass_memfault_cli_context
def upload_mcu_symbols(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload symbols for an MCU build.

    Memfault will use the Build Id in the symbol file to identify it and match
    it with data sent by that build. See https://mflt.io/symbol-file-build-ids
    for more information on enabling Build Id in your project.

    In case a Build Id cannot be found in the symbol file, a `--software-type`
    and `--software-version` must be passed. This will then be used as
    identifying information instead of a Build Id.

    Even if a Build Id is available, it is possible to pass a `--software-type`
    and `--software-version`. This will link the symbol file with the specified
    Software Version, which can be useful to easily retrieve the symbol file by
    Software Version through the Memfault UI.

    \b
    Example MCU Symbols Upload:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
            upload-mcu-symbols \\
            build/symbols.elf

    \b
    Example MCU Symbols Upload, associating it with a Software Version:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
            upload-mcu-symbols \\
            --software-type stm32-fw \\
            --software-version 1.0.0-alpha \\
            --revision 89335ffade90ff7697e2ce5238bd4c68978b6d6e \\
            build/symbols.elf
    """
    ctx.obj.update(**kwargs)

    _do_upload_or_raise(ctx, path, McuSdkElfSymbolUploader)


@cli.command(name="upload-aosp-symbols")
@click_option_concurrency
@click_argument_path
@pass_memfault_cli_context
def upload_aosp_symbols(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload symbols for an Android OS/AOSP build.

    \b
    Example AOSP Symbol Upload:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink
            upload-aosp-symbols \\
            out/target/product/generic/symbols

        Reference: https://mflt.io/android-os-symbol-files
    """
    ctx.obj.update(**kwargs)

    uploader: Type[Uploader]
    if os.path.isdir(path):
        uploader = ElfSymbolDirectoryUploader
        uploader_kwargs = dict(file_paths=walk_files(path))
    else:
        uploader = ElfSymbolUploader
        uploader_kwargs = dict()

    _do_upload_or_raise(ctx, path, uploader, **uploader_kwargs)


@cli.command(name="upload-android-app-symbols")
@click.option(
    "--build-variant",
    required=True,
    help="The build variant for which to upload the Android app symbols",
)
@click.option(
    "--package",
    required=False,
    help="The package identifier of the app. When not specified, it is read from the .apk",
)
@click.option(
    "--version-name",
    required=False,
    help="The version name of the app. When not specified, it is read from the .apk",
)
@click.option(
    "--version-code",
    required=False,
    help="The version code of the app. When not specified, it is read from the .apk",
)
@click.option(
    "--mapping-txt",
    required=False,
    help="The path to the Proguard/R8 mapping.txt file. When not specified, the default locations are searched.",
)
@click_option_concurrency
@click_argument_path
@pass_memfault_cli_context
def upload_android_app_symbols(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload symbols & R8/ProGuard mapping for an Android app build.

    Pass the root 'build' directory of the Android app as argument, for example:

    memfault upload-android-app-symbols --build-variant=release ~/my/app/build

    The command will automatically try to locate the mapping.txt and extract the
    version and package identifier from the .apk file.

    If this automatic behavior does not work in your use case, consider using
    option flags (i.e. --version-code, --version-name, --package, etc.) to specify
    the required information directly.
    """
    ctx.obj.update(**kwargs)
    uploader: Type[Uploader] = AndroidAppSymbolsUploader
    _do_upload_or_raise(ctx, path, uploader)


@cli.command(name="upload-ota-payload")
@click.option("--hardware-version", required=True)
@click.option("--software-type", required=True)
@click.option(
    "--software-version",
    required=False,
    help="Make the OTA Payload a Full Release payload for this version.",
)
@click.option(
    "--delta-from",
    required=False,
    help="Pass --delta-from FROM_VERSION make the OTA Payload a Delta Release payload from the version FROM_VERSION. Use alongside --delta-to TO_VERSION",
)
@click.option(
    "--delta-to",
    required=False,
    help="Pass --delta-to TO_VERSION to make the OTA Payload a Delta Release payload to version TO_VERSION. Use alongside --delta-from FROM_VERSION.",
)
@click.option("--notes", default="", help="Optional release notes.")
@click.option(
    "--must-pass-through",
    required=False,
    is_flag=True,
    help="When the Release is deployed to a Cohort, forces a device to update through this version even if a newer version has also been deployed to the Cohort.",
)
@click_option_revision
@click_argument_path
@pass_memfault_cli_context
def upload_ota_payload(ctx: MemfaultCliClickContext, path: str, **kwargs):
    """Upload a binary to be used for an OTA update.

    See https://mflt.io/34PyNGQ for details about 'hardware-version',
    'software-type' and 'software-version' nomenclature.

    When deployed, this is the binary that will be returned from the Memfault /latest endpoint
    which can be used for an Over The Air (OTA) update.

    \b
    Example OTA Upload:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
            upload-ota-payload \\
            --hardware-version mp \\
            --software-type stm32-fw \\
            --software-version 1.0.0-alpha \\
            --revision 89335ffade90ff7697e2ce5238bd4c68978b6d6e \\
            build/stm32-fw.bin

    \b
    Example Delta OTA Upload:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
            upload-ota-payload \\
            --hardware-version mp \\
            --software-type stm32-fw \\
            --delta-from 0.9.0 \\
            --delta-to 1.0.0-alpha \\
            --revision 89335ffade90ff7697e2ce5238bd4c68978b6d6e \\
            build/stm32-fw-delta.bin

    Reference: https://mflt.io/create-release
    """
    ctx.obj.update(**kwargs)
    ctx.check_required_either(
        {"software_version"}, {"delta_from", "delta_to"}, mutually_exclusive=True
    )
    _do_upload_or_raise(ctx, path, ReleaseArtifactUploader)


@cli.command(name="deploy-release")
@click.option("--release-version", type=str, required=False)
@click.option("--delta-from", required=False)
@click.option("--delta-to", required=False)
@click.option("--cohort", type=str, required=True)
@click.option(
    "--rollout-percent",
    type=int,
    show_default=True,
    default=100,
    help="The (randomly sampled) percentage of devices in the Cohort to rollout the release to.",
)
@pass_memfault_cli_context
def deploy_release(ctx: MemfaultCliClickContext, cohort: str, rollout_percent: int, **kwargs):
    """Publish a Release to a Cohort.

    \b
    Example Release Deployment:
    \b
        $ memfault --org-token $ORG_TOKEN \\
                   --org acme-inc --project smart-sink \\
                   deploy-release \\
                   --release-version 1.0.0-alpha \\
                   --cohort default
    """
    ctx.obj.update(**kwargs)

    authenticator = ctx.create_authenticator(OrgTokenAuthenticator, BasicAuthenticator)
    ctx.check_required_either(
        {"release_version"}, {"delta_from", "delta_to"}, mutually_exclusive=True
    )
    release_version = ctx.obj.get("release_version") or (
        ctx.obj["delta_from"],
        ctx.obj["delta_to"],
    )
    deployer = Deployer(ctx=ctx, authenticator=authenticator)
    deployer.deploy(cohort=cohort, release_version=release_version, rollout_percent=rollout_percent)


def _do_post_chunks_or_raise(ctx: MemfaultCliClickContext, chunks: List[bytes]):
    authenticator = ctx.create_authenticator(ProjectKeyAuthenticator)

    MemfaultChunk(ctx, authenticator).batch_post(chunks)
    click.echo("Success")


@cli.command("post-chunk")
@click.option("--device-serial", show_default=True, default="TESTSERIAL")
@click.option(
    "--encoding",
    type=click.Choice(["hex", "base64", "bin", "sdk_data_export"]),
    required=True,
    help="The format DATA is encoded in.",
)
@click.argument("data")
@pass_memfault_cli_context
def post_chunk(ctx: MemfaultCliClickContext, encoding, data, **kwargs):
    """Sends data generated by the memfault-firmware-sdk ("chunks") to the Memfault cloud.

    The command can operate on binary data encoded in the following formats:

    \b
    1. Hex String:
      memfault --project-key ${YOUR_PROJECT_KEY} post-chunk --encoding hex 0802a702010301076a5445535453455249414c0a6d746573742d736f667477617265096a312e302e302d74657374066d746573742d686172647761726504a101a1726368756e6b5f746573745f737563636573730131e4

    \b
    2. Base64 Encoded String
      memfault --project-key ${YOUR_PROJECT_KEY} post-chunk --encoding base64 CAKnAgEDAQdqVEVTVFNFUklBTAptdGVzdC1zb2Z0d2FyZQlqMS4wLjAtdGVzdAZtdGVzdC1oYXJkd2FyZQShAaFyY2h1bmtfdGVzdF9zdWNjZXNzATHk

    \b
    3. Binary File
      memfault --project-key ${YOUR_PROJECT_KEY} post-chunk --encoding bin chunk_v2_single_chunk_msg.bin

    \b
    4. memfault-firmware-sdk data export
      memfault --project-key ${YOUR_PROJECT_KEY} post-chunk --encoding sdk_data_export data_export.txt

    Reference: https://mflt.io/chunk-api-integration
    """

    ctx.obj.update(**kwargs)
    if encoding == "hex":
        chunks = [bytes.fromhex(data)]
    elif encoding == "base64":
        chunks = [b64decode(data)]
    elif encoding == "bin":
        with click.open_file(data, "rb") as f:
            chunks = [f.read()]
    elif encoding == "sdk_data_export":
        with click.open_file(data, "r") as f:
            exported_data = f.read()
        chunks = MemfaultChunk.extract_exported_chunks(exported_data)
        if len(chunks) == 0:
            raise click.exceptions.UsageError(f"No Memfault chunks found in {data}")
        click.echo(f"Found {len(chunks)} Chunks. Sending Data ...")

    _do_post_chunks_or_raise(ctx, chunks)


@cli.command("completion")
@click.option(
    "--shell",
    type=click.Choice(["bash", "zsh", "fish"]),
    required=True,
    help="select shell type for completion script",
)
@click.pass_context
def completion(ctx, shell: str):
    """
    Generate shell completion script for memfault.

    \b
    Example:
    \b
        $ memfault completion --shell bash > ~/.memfault_completion.bash
        $ echo 'source ~/.memfault_completion.bash' >> ~/.bashrc

    """

    # re-invoke the main cli entrance point after setting the Click completion
    # environment variable. it's necessary to do this via the Click ctx so that
    # it works properly when invoked from pytest or normal cli usage.

    cmd = "memfault"
    os.environ[f"_{cmd.upper()}_COMPLETE"] = f"{shell}_source"
    ctx.command.main(prog_name="memfault")


def main():
    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
    cli(auto_envvar_prefix="MEMFAULT")


if __name__ == "__main__":
    main()
